import os
import asyncio
from metaapi_cloud_sdk import MetaApi
from datetime import datetime

token = os.getenv('TOKEN') or '<put in your token here>'
account_id = os.getenv('ACCOUNT_ID') or '<put in your account id here>'
symbol = os.getenv('SYMBOL') or 'EURUSD'
domain = os.getenv('DOMAIN') or 'agiliumtrade.agiliumtrade.ai'


async def retrieve_historical_ticks():
    api = MetaApi(token, {'domain': domain})
    try:
        account = await api.metatrader_account_api.get_account(account_id)

        #  wait until account is deployed and connected to broker
        print('Deploying account')
        if account.state != 'DEPLOYED':
            await account.deploy()
        else:
            print('Account already deployed')
        print('Waiting for API server to connect to broker (may take couple of minutes)')
        if account.connection_status != 'CONNECTED':
            await account.wait_connected()

        # retrieve last 10K ticks
        pages = 10
        print(f'Downloading {pages}K ticks for {symbol} starting from 7 days ago')
        started_at = datetime.now().timestamp()
        start_time = datetime.fromtimestamp(datetime.now().timestamp() - 7 * 24 * 60 * 60)
        offset = 0
        ticks = None
        for i in range(pages):
            # the API to retrieve historical market data is currently available for G1 only
            # historical ticks can be retrieved from MT5 only
            ticks = await account.get_historical_ticks(symbol, start_time, offset)
            print(f'Downloaded {len(ticks) if ticks else 0} historical ticks for {symbol}')
            if ticks and len(ticks):
                start_time = ticks[-1]['time']
                offset = 0
                while offset < len(ticks) and ticks[-1 - offset]['time'].timestamp() == start_time.timestamp():
                    offset += 1
                print(f'Last tick time is {start_time}, offset is {offset}')
        if ticks:
            print('Last tick is', ticks[-1])
        print(f'Took {(datetime.now().timestamp() - started_at) * 1000}ms')

    except Exception as err:
        print(api.format_error(err))

asyncio.run(retrieve_historical_ticks())



import asyncio
from random import uniform
from ..errorHandler import TooManyRequestsException
from ...metaApi.models import date, format_error
from datetime import datetime
from typing import List


class SubscriptionManager:
    """Subscription manager to handle account subscription logic."""

    def __init__(self, websocket_client):
        """Inits the subscription manager.

        Args:
            websocket_client: Websocket client to use for sending requests.
        """
        self._websocketClient = websocket_client
        self._subscriptions = {}
        self._awaitingResubscribe = {}

    def is_account_subscribing(self, account_id: str, instance_number: int = None):
        """Returns whether an account is currently subscribing.

        Args:
            account_id: Id of the MetaTrader account.
            instance_number: Instance index number.
        """
        if instance_number is not None:
            return account_id + ':' + str(instance_number) in self._subscriptions.keys()
        else:
            for key in self._subscriptions.keys():
                if key.startswith(account_id):
                    return True
            return False

    def is_disconnected_retry_mode(self, account_id: str, instance_number: int):
        """Returns whether an instance is in disconnected retry mode.

        Args:
            account_id: Id of the MetaTrader account.
            instance_number: Instance index number.
        """
        instance_id = account_id + ':' + str(instance_number or 0)
        return self._subscriptions[instance_id]['isDisconnectedRetryMode'] if instance_id in \
            self._subscriptions else False

    async def subscribe(self, account_id: str, instance_number: int = None, is_disconnected_retry_mode=False):
        """Schedules to send subscribe requests to an account until cancelled.

        Args:
            account_id: Id of the MetaTrader account.
            instance_number: Instance index number.
            is_disconnected_retry_mode: Whether to start subscription in disconnected retry mode. Subscription
                task in disconnected mode will be immediately replaced when the status packet is received.
        """
        instance_id = account_id + ':' + str(instance_number or 0)
        if instance_id not in self._subscriptions:
            self._subscriptions[instance_id] = {
                'shouldRetry': True,
                'task': None,
                'wait_task': None,
                'future': None,
                'isDisconnectedRetryMode': is_disconnected_retry_mode
            }
            subscribe_retry_interval_in_seconds = 3
            while self._subscriptions[instance_id]['shouldRetry']:
                async def subscribe_task():
                    try:
                        await self._websocketClient.subscribe(account_id, instance_number)
                    except TooManyRequestsException as err:
                        socket_instance_index = self._websocketClient.socket_instances_by_accounts[account_id]
                        if err.metadata['type'] == 'LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_USER':
                            print(format_error(err))
                        if err.metadata['type'] in ['LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_USER',
                                                    'LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_SERVER',
                                                    'LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_USER_PER_SERVER']:
                            del self._websocketClient.socket_instances_by_accounts[account_id]
                            asyncio.create_task(self._websocketClient.lock_socket_instance(socket_instance_index,
                                                                                           err.metadata))
                        else:
                            nonlocal subscribe_retry_interval_in_seconds
                            retry_time = date(err.metadata['recommendedRetryTime']).timestamp()
                            if datetime.now().timestamp() + subscribe_retry_interval_in_seconds < retry_time:
                                await asyncio.sleep(retry_time - datetime.now().timestamp() -
                                                    subscribe_retry_interval_in_seconds)
                    except Exception as err:
                        pass

                self._subscriptions[instance_id]['task'] = asyncio.create_task(subscribe_task())
                await asyncio.wait({self._subscriptions[instance_id]['task']})
                if not self._subscriptions[instance_id]['shouldRetry']:
                    break
                retry_interval = subscribe_retry_interval_in_seconds
                subscribe_retry_interval_in_seconds = min(subscribe_retry_interval_in_seconds * 2, 300)
                subscribe_future = asyncio.Future()

                async def subscribe_task():
                    await asyncio.sleep(retry_interval)
                    subscribe_future.set_result(True)

                self._subscriptions[instance_id]['wait_task'] = asyncio.create_task(subscribe_task())
                self._subscriptions[instance_id]['future'] = subscribe_future
                result = await self._subscriptions[instance_id]['future']
                self._subscriptions[instance_id]['future'] = None
                if not result:
                    break
            del self._subscriptions[instance_id]

    def cancel_subscribe(self, instance_id: str):
        """Cancels active subscription tasks for an instance id.

        Args:
            instance_id: Instance id to cancel subscription task for.
        """
        if instance_id in self._subscriptions:
            subscription = self._subscriptions[instance_id]
            if subscription['future'] and not subscription['future'].done():
                subscription['future'].set_result(False)
                subscription['wait_task'].cancel()
            if subscription['task']:
                subscription['task'].cancel()
            subscription['shouldRetry'] = False

    def cancel_account(self, account_id):
        """Cancels active subscription tasks for an account.

        Args:
            account_id: Account id to cancel subscription tasks for.
        """
        for instance_id in list(filter(lambda key: key.startswith(account_id), self._subscriptions.keys())):
            self.cancel_subscribe(instance_id)

    def on_timeout(self, account_id: str, instance_number: int = None):
        """Invoked on account timeout.

        Args:
            account_id: Id of the MetaTrader account.
            instance_number: Instance index number.
        """
        if account_id in self._websocketClient.socket_instances_by_accounts and \
                self._websocketClient.connected(self._websocketClient.socket_instances_by_accounts[account_id]):
            asyncio.create_task(self.subscribe(account_id, instance_number, is_disconnected_retry_mode=True))

    async def on_disconnected(self, account_id: str, instance_number: int = None):
        """Invoked when connection to MetaTrader terminal terminated.

        Args:
            account_id: Id of the MetaTrader account.
            instance_number: Instance index number.
        """
        await asyncio.sleep(uniform(1, 5))
        if account_id in self._websocketClient.socket_instances_by_accounts:
            asyncio.create_task(self.subscribe(account_id, instance_number, is_disconnected_retry_mode=True))

    def on_reconnected(self, socket_instance_index: int, reconnect_account_ids: List[str]):
        """Invoked when connection to MetaApi websocket API restored after a disconnect.

        Args:
            socket_instance_index: Socket instance index.
            reconnect_account_ids: Account ids to reconnect.
        """

        async def wait_resubscribe(account_id):
            try:
                if account_id not in self._awaitingResubscribe:
                    self._awaitingResubscribe[account_id] = True
                    while self.is_account_subscribing(account_id):
                        await asyncio.sleep(1)
                    if account_id in self._awaitingResubscribe:
                        del self._awaitingResubscribe[account_id]
                    await asyncio.sleep(uniform(0, 5))
                    asyncio.create_task(self.subscribe(account_id))
            except Exception as err:
                print(f'[{datetime.now().isoformat()}] Account {account_id} resubscribe task failed',
                      format_error(err))

        try:
            socket_instances_by_accounts = self._websocketClient.socket_instances_by_accounts
            for instance_id in self._subscriptions.keys():
                account_id = instance_id.split(':')[0]
                if account_id in socket_instances_by_accounts and \
                        socket_instances_by_accounts[account_id] == socket_instance_index:
                    self.cancel_subscribe(instance_id)

            for account_id in reconnect_account_ids:
                asyncio.create_task(wait_resubscribe(account_id))
        except Exception as err:
            print(f'[{datetime.now().isoformat()}] Failed to process subscribe manager reconnected event',
                  format_error(err))



from typing import Dict
from collections import deque
from datetime import datetime
import asyncio
from typing_extensions import TypedDict
from typing import Optional, List
from functools import reduce
from ..timeoutException import TimeoutException
from ..optionsValidator import OptionsValidator
import math


class SynchronizationThrottlerOpts(TypedDict):
    """Options for synchronization throttler."""
    maxConcurrentSynchronizations: Optional[int]
    """Amount of maximum allowed concurrent synchronizations."""
    queueTimeoutInSeconds: Optional[float]
    """Allowed time for a synchronization in queue."""
    synchronizationTimeoutInSeconds: Optional[float]
    """Time after which a synchronization slot is freed to be used by another synchronization."""


class SynchronizationThrottler:
    """Synchronization throttler used to limit the amount of concurrent synchronizations to prevent application
    from being overloaded due to excessive number of synchronisation responses being sent."""

    def __init__(self, client, socket_instance_index: int, opts: SynchronizationThrottlerOpts = None):
        """Inits the synchronization throttler.

        Args:
            client: Websocket client.
            socket_instance_index: Index of socket instance that uses the throttler.
            opts: Synchronization throttler options.
        """
        validator = OptionsValidator()
        opts: SynchronizationThrottlerOpts = opts or {}
        self._maxConcurrentSynchronizations = validator.validate_non_zero(
            opts['maxConcurrentSynchronizations'] if 'maxConcurrentSynchronizations' in opts else None, 15,
            'synchronizationThrottler.maxConcurrentSynchronizations')
        self._queueTimeoutInSeconds = validator.validate_non_zero(
            opts['queueTimeoutInSeconds'] if 'queueTimeoutInSeconds' in opts else None, 300,
            'synchronizationThrottler.queueTimeoutInSeconds')
        self._synchronizationTimeoutInSeconds = validator.validate_non_zero(
            opts['synchronizationTimeoutInSeconds'] if 'synchronizationTimeoutInSeconds' in opts else None, 10,
            'synchronizationThrottler.synchronizationTimeoutInSeconds')
        self._client = client
        self._socketInstanceIndex = socket_instance_index
        self._synchronizationIds = {}
        self._accountsBySynchronizationIds = {}
        self._synchronizationQueue = deque([])
        self._removeOldSyncIdsInterval = None
        self._processQueueInterval = None

    def start(self):
        """Initializes the synchronization throttler."""
        async def remove_old_sync_ids_interval():
            while True:
                await self._remove_old_sync_ids_job()
                await asyncio.sleep(1)

        async def process_queue_interval():
            while True:
                await self._process_queue_job()
                await asyncio.sleep(1)

        if not self._removeOldSyncIdsInterval:
            self._removeOldSyncIdsInterval = asyncio.create_task(remove_old_sync_ids_interval())
            self._processQueueInterval = asyncio.create_task(process_queue_interval())

    def stop(self):
        """Deinitializes the throttler."""
        if self._removeOldSyncIdsInterval:
            self._removeOldSyncIdsInterval.cancel()
            self._removeOldSyncIdsInterval = None
        if self._processQueueInterval:
            self._processQueueInterval.cancel()
            self._processQueueInterval = None

    async def _remove_old_sync_ids_job(self):
        now = datetime.now().timestamp()
        for key in list(self._synchronizationIds.keys()):
            if (now - self._synchronizationIds[key]) > self._synchronizationTimeoutInSeconds:
                del self._synchronizationIds[key]
        while len(self._synchronizationQueue) and \
                (datetime.now().timestamp() - self._synchronizationQueue[0]['queueTime']) > self._queueTimeoutInSeconds:
            self._remove_from_queue(self._synchronizationQueue[0]['synchronizationId'], 'timeout')
        self._advance_queue()
        await asyncio.sleep(1)

    def update_synchronization_id(self, synchronization_id: str):
        """Fills a synchronization slot with synchronization id.

        Args:
            synchronization_id: Synchronization id.
        """
        if synchronization_id in self._accountsBySynchronizationIds:
            self._synchronizationIds[synchronization_id] = datetime.now().timestamp()

    @property
    def synchronizing_accounts(self) -> List[str]:
        """Returns the list of currently synchronizing account ids."""
        synchronizing_accounts = []
        for key in self._synchronizationIds:
            account_data = self._accountsBySynchronizationIds[key] if key in \
                                                                      self._accountsBySynchronizationIds else None
            if account_data and (account_data['accountId'] not in synchronizing_accounts):
                synchronizing_accounts.append(account_data['accountId'])
        return synchronizing_accounts

    @property
    def active_synchronization_ids(self) -> List[str]:
        """Returns the list of currently active synchronization ids."""
        return list(self._accountsBySynchronizationIds.keys())

    @property
    def max_concurrent_synchronizations(self) -> int:
        """Returns the amount of maximum allowed concurrent synchronizations."""
        calculated_max = max(math.ceil(len(self._client.subscribed_account_ids(self._socketInstanceIndex)) / 10), 1)
        return min(calculated_max, self._maxConcurrentSynchronizations)

    @property
    def is_synchronization_available(self) -> bool:
        """Whether there are free slots for synchronization requests."""

        def reducer_func(acc, socket_instance):
            return acc + len(socket_instance['synchronizationThrottler'].synchronizing_accounts)

        if reduce(reducer_func, self._client.socket_instances, 0) >= self._maxConcurrentSynchronizations:
            return False
        return len(self.synchronizing_accounts) < self.max_concurrent_synchronizations

    def remove_id_by_parameters(self, account_id: str, instance_index: int = None, host: str = None):
        """Removes synchronizations from queue and from the list by parameters.

        Args:
            account_id: Account id.
            instance_index: Account instance index.
            host: Account host name.
        """
        for key in list(self._accountsBySynchronizationIds.keys()):
            if self._accountsBySynchronizationIds[key]['accountId'] == account_id and \
                    self._accountsBySynchronizationIds[key]['instanceIndex'] == instance_index and \
                    self._accountsBySynchronizationIds[key]['host'] == host:
                self.remove_synchronization_id(key)

    def remove_synchronization_id(self, synchronization_id: str):
        """Removes synchronization id from slots and removes ids for the same account from the queue.

        Args:
            synchronization_id: Synchronization id.
        """
        if synchronization_id in self._accountsBySynchronizationIds:
            account_id = self._accountsBySynchronizationIds[synchronization_id]['accountId']
            instance_index = self._accountsBySynchronizationIds[synchronization_id]['instanceIndex']
            host = self._accountsBySynchronizationIds[synchronization_id]['host']
            for key in list(self._accountsBySynchronizationIds.keys()):
                if self._accountsBySynchronizationIds[key]['accountId'] == account_id and \
                        self._accountsBySynchronizationIds[key]['instanceIndex'] == instance_index and \
                        self._accountsBySynchronizationIds[key]['host'] == host:
                    self._remove_from_queue(key, 'cancel')
                    del self._accountsBySynchronizationIds[key]
        if synchronization_id in self._synchronizationIds:
            del self._synchronizationIds[synchronization_id]
        self._advance_queue()

    def on_disconnect(self):
        """Clears synchronization ids on disconnect."""
        for synchronization in self._synchronizationQueue:
            if not synchronization['promise'].done():
                synchronization['promise'].set_result('cancel')
        self._synchronizationIds = {}
        self._accountsBySynchronizationIds = {}
        self._synchronizationQueue = []
        self.stop()
        self.start()

    def _advance_queue(self):
        index = 0
        while self.is_synchronization_available and len(self._synchronizationQueue) and index < \
                len(self._synchronizationQueue):
            queue_item = self._synchronizationQueue[index]
            if not queue_item['promise'].done():
                queue_item['promise'].set_result('synchronize')
                self.update_synchronization_id(queue_item['synchronizationId'])
            index += 1

    def _remove_from_queue(self, synchronization_id: str, result: str):
        for i in range(len(self._synchronizationQueue)):
            sync_item = self._synchronizationQueue[i]
            if sync_item['synchronizationId'] == synchronization_id and not sync_item['promise'].done():
                sync_item['promise'].set_result(result)
        self._synchronizationQueue = deque(filter(lambda item: item['synchronizationId'] != synchronization_id,
                                                  self._synchronizationQueue))

    async def _process_queue_job(self):
        while len(self._synchronizationQueue):
            queue_item = self._synchronizationQueue[0]
            await queue_item['promise']
            if len(self._synchronizationQueue) and self._synchronizationQueue[0]['synchronizationId'] == \
                    queue_item['synchronizationId']:
                self._synchronizationQueue.popleft()

    async def schedule_synchronize(self, account_id: str, request: Dict):
        """Schedules to send a synchronization request for account.

        Args:
            account_id: Account id.
            request: Request to send.
        """
        synchronization_id = request['requestId']
        for key in list(self._accountsBySynchronizationIds.keys()):
            if self._accountsBySynchronizationIds[key]['accountId'] == account_id and \
                     self._accountsBySynchronizationIds[key]['instanceIndex'] == \
                    (request['instanceIndex'] if 'instanceIndex' in request else None):
                self.remove_synchronization_id(key)
        self._accountsBySynchronizationIds[synchronization_id] = {
            'accountId': account_id, 'instanceIndex': request['instanceIndex'] if 'instanceIndex' in request else None,
            'host': request['host'] if 'host' in request else None
        }
        if not self.is_synchronization_available:
            request_resolve = asyncio.Future()
            self._synchronizationQueue.append({
                'synchronizationId': synchronization_id,
                'promise': request_resolve,
                'queueTime': datetime.now().timestamp()
            })
            result = await request_resolve
            if result == 'cancel':
                return False
            elif result == 'timeout':
                raise TimeoutException(f'Account {account_id} synchronization {synchronization_id} timed out in '
                                       f'synchronization queue')
        self.update_synchronization_id(synchronization_id)
        await self._client._rpc_request(account_id, request)
        return True



from .. import MetaApi
from ..metaApi.models import format_date
from socketio import AsyncServer
from aiohttp import web
import pytest
import asyncio
from asyncio import sleep
from mock import patch, AsyncMock, MagicMock
from datetime import datetime, timedelta

sio: AsyncServer = None
client_sid: str = ''
api: MetaApi = None
account_information = {
    'broker': 'True ECN Trading Ltd',
    'currency': 'USD',
    'server': 'ICMarketsSC-Demo',
    'balance': 7319.9,
    'equity': 7306.649913200001,
    'margin': 184.1,
    'freeMargin': 7120.22,
    'leverage': 100,
    'marginLevel': 3967.58283542
}
errors = [
    {
        "id": 1,
        "error": "TooManyRequestsError",
        "message": "One user can connect to one server no more than 300 accounts. Current number of connected "
                   "accounts 300. For more information see https://metaapi.cloud/docs/client/rateLimiting/",
        "metadata": {
            "maxAccountsPerUserPerServer": 300,
            "accountsCount":  300,
            "recommendedRetryTime": format_date(datetime.now() + timedelta(seconds=20)),
            "type": "LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_USER_PER_SERVER"
        }
    },
    {
        "id": 1,
        "error": "TooManyRequestsError",
        "message": "You have used all your account subscriptions quota. You have 50 account subscriptions available "
                   "and have used 50 subscriptions. Please deploy more accounts to get more subscriptions. For more "
                   "information see https://metaapi.cloud/docs/client/rateLimiting/",
        "metadata": {
            "maxAccountsPerUser":  50,
            "accountsCount": 50,
            "recommendedRetryTime": format_date(datetime.now() + timedelta(seconds=20)),
            "type": "LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_USER"
        }
    },
    {
        "id": 1,
        "error": "TooManyRequestsError",
        "message": "You can not subscribe to more accounts on this connection because server is out of capacity. "
                   "Please establish a new connection with a different client-id header value to switch to a "
                   "different server. For more information see https://metaapi.cloud/docs/client/rateLimiting/",
        "metadata": {
            "changeClientIdHeader": True,
            "recommendedRetryTime": format_date(datetime.now() + timedelta(seconds=20)),
            "type": "LIMIT_ACCOUNT_SUBSCRIPTIONS_PER_SERVER"
        }
    }
]


class FakeServer:

    def __init__(self):
        self.app = web.Application()
        self.sio: AsyncServer = None
        self.runner = None
        self.stopped = False
        self.status_tasks = {}
        self.client_ids = []

    async def authenticate(self, data, host='ps-mpa-0'):
        await self.sio.emit('synchronization', {'type': 'authenticated', 'accountId': data['accountId'],
                                                'instanceIndex': 0, 'replicas': 1, 'host': host})

    async def emit_status(self, account_id: str, host='ps-mpa-0'):
        packet = {'connected': True, 'authenticated': True, 'instanceIndex': 0, 'type': 'status',
                  'healthStatus': {'rpcApiHealthy': True}, 'replicas': 1, 'host': host,
                  'connectionId': account_id, 'accountId': account_id}
        await self.sio.emit('synchronization', packet)

    async def create_status_task(self, account_id: str, host='ps-mpa-0'):
        while True:
            await self.emit_status(account_id, host)
            await sleep(1)

    def delete_status_task(self, account_id: str):
        if account_id in self.status_tasks:
            self.status_tasks[account_id].cancel()
            del self.status_tasks[account_id]

    async def respond_account_information(self, data):
        await self.sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                         'requestId': data['requestId'], 'accountInformation': account_information})

    async def sync_account(self, data, host='ps-mpa-0'):
        await self.sio.emit('synchronization', {'type': 'synchronizationStarted', 'accountId': data['accountId'],
                                                'instanceIndex': 0, 'synchronizationId': data['requestId'],
                                                'host': host})
        await sleep(0.1)
        await self.sio.emit('synchronization', {'type': 'accountInformation', 'accountId': data['accountId'],
                                                'accountInformation': account_information, 'instanceIndex': 0,
                                                'host': host})
        await self.sio.emit('synchronization', {'type': 'specifications', 'accountId': data['accountId'],
                                                'specifications': [], 'instanceIndex': 0, 'host': host})
        await self.sio.emit('synchronization', {'type': 'orderSynchronizationFinished',
                                                'accountId': data['accountId'], 'instanceIndex': 0,
                                                'synchronizationId': data['requestId'], 'host': host})
        await sleep(0.1)
        await self.sio.emit('synchronization', {'type': 'dealSynchronizationFinished',
                                                'accountId': data['accountId'], 'instanceIndex': 0,
                                                'synchronizationId': data['requestId'], 'host': host})

    async def respond(self, data):
        await self.sio.emit('response', {'type': 'response', 'accountId': data['accountId'],
                                         'requestId': data['requestId']})

    async def emit_error(self, data, error_index, retry_after_seconds):
        error = errors[error_index]
        error['metadata']['recommendedRetryTime'] = format_date(datetime.now() +
                                                                timedelta(seconds=retry_after_seconds))
        await sio.emit('processingError', {**error, 'requestId': data['requestId']})

    def enable_sync(self):
        @self.sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe':
                await sleep(0.2)
                await self.respond(data)
                self.status_tasks[data['accountId']] = asyncio.create_task(self.create_status_task(data['accountId']))
                await self.authenticate(data)
            elif data['type'] == 'synchronize':
                await self.respond(data)
                await self.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await self.respond(data)
            elif data['type'] == 'getAccountInformation':
                await self.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                self.delete_status_task(data['accountId'])
                await self.respond(data)

    def disable_sync(self):
        @self.sio.on('request')
        async def on_request(sid, data):
            return False

    async def start(self):
        global sio
        sio = AsyncServer(async_mode='aiohttp')
        self.sio = sio

        @sio.event
        async def connect(sid, environ):
            self.client_ids.append(environ['aiohttp.request'].headers['Client-Id'])
            global client_sid
            client_sid = sid
            await sio.emit('response', {'type': 'response'})

        self.enable_sync()
        sio.attach(self.app, socketio_path='ws')
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, 'localhost', 8080)
        await site.start()

    async def stop(self):
        if not self.stopped:
            self.stopped = True
            await self.runner.cleanup()


fake_server: FakeServer = None


@pytest.fixture(autouse=True, params=[True, False])
async def run_around_tests(request):
    global fake_server
    fake_server = FakeServer()
    await fake_server.start()
    global api
    api = MetaApi('token', {'application': 'application', 'domain': 'project-stock.agiliumlabs.cloud',
                            'useSharedClientApi': True,
                            'eventProcessing': {'sequentialProcessing': request.param},
                            'requestTimeout': 3, 'retryOpts': {'retries': 3, 'minDelayInSeconds': 0.1,
                                                               'maxDelayInSeconds': 0.5,
                                                               'subscribeCooldownInSeconds': 6}})

    async def side_effect_get_account(account_id):
        return {
            '_id': account_id,
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud',
            'accessToken': '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        }

    api.metatrader_account_api._metatraderAccountClient.get_account = side_effect_get_account
    api._metaApiWebsocketClient.set_url('http://localhost:8080')
    await api._metaApiWebsocketClient.connect()
    api._metaApiWebsocketClient._resolved = True
    yield
    tasks = [task for task in asyncio.all_tasks() if task is not
             asyncio.tasks.current_task()]
    list(map(lambda task: task.cancel(), tasks))
    await fake_server.stop()


class TestSyncStability:
    @pytest.mark.asyncio
    async def test_sync(self):
        """Should synchronize account"""
        account = await api.metatrader_account_api.get_account('accountId')
        connection = await account.connect()
        await connection.wait_synchronized({'timeoutInSeconds': 10})
        response = await connection.get_account_information()
        assert response == account_information
        assert connection.synchronized and connection.terminal_state.connected and \
               connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_socket_disconnect(self):
        """Should reconnect on server socket crash."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            await sio.disconnect(client_sid)
            await sleep(0.5)
            response = await connection.get_account_information()
            assert response == account_information
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_set_disconnected(self):
        """Should set state to disconnected on timeout."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            fake_server.delete_status_task('accountId')

            @sio.event
            async def connect(sid, environ):
                return False

            await sio.disconnect(client_sid)
            await sleep(1.2)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_resubscribe_on_timeout(self):
        """Should resubscribe on timeout."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            fake_server.status_tasks['accountId'].cancel()
            await sleep(1.5)
            response = await connection.get_account_information()
            assert response == account_information
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_subscribe_with_late_response(self):
        """Should synchronize if subscribe response arrives after synchronization."""

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe':
                await sleep(0.2)
                fake_server.status_tasks[data['accountId']] = \
                    asyncio.create_task(fake_server.create_status_task(data['accountId']))
                await fake_server.authenticate(data)
                await sleep(0.4)
                await fake_server.respond(data)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)

        account = await api.metatrader_account_api.get_account('accountId')
        connection = await account.connect()
        await connection.wait_synchronized({'timeoutInSeconds': 10})
        response = await connection.get_account_information()
        assert response == account_information
        assert connection.synchronized and connection.terminal_state.connected and \
               connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_wait_redeploy(self):
        """Should wait until account is redeployed after disconnect."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            fake_server.delete_status_task('accountId')
            fake_server.disable_sync()
            await sleep(0.05)
            await sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId', 'host': 'ps-mpa-0',
                                               'instanceIndex': 0})
            await sleep(0.4)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker
            await sleep(4)
            fake_server.enable_sync()
            await sleep(0.4)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker
            await sleep(4)
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_resubscribe_on_status_packet(self):
        """Should resubscribe immediately after disconnect on status packet."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            fake_server.delete_status_task('accountId')
            await sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId',
                                               'host': 'ps-mpa-0', 'instanceIndex': 0})
            fake_server.disable_sync()
            await sleep(0.4)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker
            await sleep(4)
            fake_server.enable_sync()
            await fake_server.emit_status('accountId')
            await sleep(0.4)
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_resubscribe_on_reconnected_after_connection_closed(self):
        """Should resubscribe other accounts after one of connections is closed"""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            account2 = await api.metatrader_account_api.get_account('accountId2')
            connection2 = await account2.connect()
            await connection2.wait_synchronized({'timeoutInSeconds': 10})
            account3 = await api.metatrader_account_api.get_account('accountId3')
            connection3 = await account3.connect()
            await connection3.wait_synchronized({'timeoutInSeconds': 10})
            await connection.close()
            fake_server.delete_status_task('accountId2')
            fake_server.delete_status_task('accountId3')
            fake_server.disable_sync()
            await sio.disconnect(client_sid)
            await sleep(2)
            fake_server.enable_sync()
            await sio.disconnect(client_sid)
            await sleep(3)
            assert not connection.synchronized
            assert connection2.synchronized and connection2.terminal_state.connected and \
                   connection2.terminal_state.connected_to_broker
            assert connection3.synchronized and connection3.terminal_state.connected and \
                   connection3.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_429_per_user_limit_subscriptions(self):
        """Should limit subscriptions during per user 429 error."""
        subscribed_accounts = {}

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal subscribed_accounts
            if data['type'] == 'subscribe':
                if len(subscribed_accounts.keys()) < 2:
                    subscribed_accounts[data['accountId']] = True
                    await sleep(0.2)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
                else:
                    await fake_server.emit_error(data, 1, 2)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                del subscribed_accounts[data['accountId']]
                await fake_server.respond(data)

        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 3})
            account2 = await api.metatrader_account_api.get_account('accountId2')
            connection2 = await account2.connect()
            await connection2.wait_synchronized({'timeoutInSeconds': 3})
            account3 = await api.metatrader_account_api.get_account('accountId3')
            connection3 = await account3.connect()
            try:
                await connection3.wait_synchronized({'timeoutInSeconds': 3})
                raise Exception('TimeoutException expected')
            except Exception as err:
                assert err.__class__.__name__ == 'TimeoutException'
            await connection2.close()
            await sleep(2)
            assert connection3.synchronized

    @pytest.mark.asyncio
    async def test_429_per_user_retry_after_time(self):
        """Should wait for retry time after per user 429 error."""
        request_timestamp = 0
        subscribed_accounts = {}

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal subscribed_accounts
            nonlocal request_timestamp
            if data['type'] == 'subscribe':
                if len(subscribed_accounts.keys()) < 2 or (request_timestamp != 0 and datetime.now().timestamp() - 2 >
                                                           request_timestamp):
                    subscribed_accounts[data['accountId']] = True
                    await sleep(0.2)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
                else:
                    request_timestamp = datetime.now().timestamp()
                    await fake_server.emit_error(data, 1, 3)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                del subscribed_accounts[data['accountId']]
                await fake_server.respond(data)

        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 3})
            account2 = await api.metatrader_account_api.get_account('accountId2')
            connection2 = await account2.connect()
            await connection2.wait_synchronized({'timeoutInSeconds': 3})
            account3 = await api.metatrader_account_api.get_account('accountId3')
            connection3 = await account3.connect()
            try:
                await connection3.wait_synchronized({'timeoutInSeconds': 3})
                raise Exception('TimeoutException expected')
            except Exception as err:
                assert err.__class__.__name__ == 'TimeoutException'
            await sleep(2)
            assert not connection3.synchronized
            await sleep(2.5)
            assert connection3.synchronized

    @pytest.mark.asyncio
    async def test_429_per_server_retry_after_time(self):
        """Should wait for retry time after per server 429 error."""
        sid_by_accounts = {}
        request_timestamp = 0

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal request_timestamp
            if data['type'] == 'subscribe':
                if len(list(filter(lambda account_sid: account_sid == sid, sid_by_accounts.values()))) >= 2 and \
                        (request_timestamp == 0 or datetime.now().timestamp() - 2 < request_timestamp):
                    request_timestamp = datetime.now().timestamp()
                    await fake_server.emit_error(data, 2, 2)
                else:
                    sid_by_accounts[data['accountId']] = sid
                    await sleep(0.2)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)

        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 3})
            account2 = await api.metatrader_account_api.get_account('accountId2')
            connection2 = await account2.connect()
            await connection2.wait_synchronized({'timeoutInSeconds': 3})
            account3 = await api.metatrader_account_api.get_account('accountId3')
            connection3 = await account3.connect()
            await connection3.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] == sid_by_accounts['accountId2'] != sid_by_accounts['accountId3']
            await sleep(2)
            account4 = await api.metatrader_account_api.get_account('accountId4')
            connection4 = await account4.connect()
            await connection4.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] == sid_by_accounts['accountId4']

    @pytest.mark.asyncio
    async def test_429_per_server_reconnect(self):
        """Should reconnect after per server 429 error if connection has no subscribed accounts."""
        sids = []

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe':
                sids.append(sid)
                if len(sids) == 1:
                    await fake_server.emit_error(data, 2, 2)
                else:
                    await sleep(0.2)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                fake_server.delete_status_task(data['accountId'])
                await fake_server.respond(data)

        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 3})
            assert sids[0] != sids[1]

    @pytest.mark.asyncio
    async def test_429_per_server_unsubscribe(self):
        """Should free a subscribe slot on unsubscribe after per server 429 error."""
        sid_by_accounts = {}

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe':
                if len(list(filter(lambda account_sid: account_sid == sid, sid_by_accounts.values()))) >= 2:
                    await fake_server.emit_error(data, 2, 200)
                else:
                    sid_by_accounts[data['accountId']] = sid
                    await sleep(0.2)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                del sid_by_accounts[data['accountId']]
                await fake_server.respond(data)

        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 3})
            account2 = await api.metatrader_account_api.get_account('accountId2')
            connection2 = await account2.connect()
            await connection2.wait_synchronized({'timeoutInSeconds': 3})
            account3 = await api.metatrader_account_api.get_account('accountId3')
            connection3 = await account3.connect()
            await connection3.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] == sid_by_accounts['accountId2'] != sid_by_accounts['accountId3']
            await connection2.close()
            account4 = await api.metatrader_account_api.get_account('accountId4')
            connection4 = await account4.connect()
            await connection4.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] == sid_by_accounts['accountId4']

    @pytest.mark.asyncio
    async def test_429_per_server_per_user_retry_after_time(self):
        """Should wait for retry time after per server per user 429 error."""
        sid_by_accounts = {}
        request_timestamp = 0

        @sio.on('request')
        async def on_request(sid, data):
            nonlocal request_timestamp
            if data['type'] == 'subscribe':
                if len(list(filter(lambda account_sid: account_sid == sid, sid_by_accounts.values()))) >= 2 and \
                        (request_timestamp == 0 or datetime.now().timestamp() - 2 < request_timestamp):
                    request_timestamp = datetime.now().timestamp()
                    await fake_server.emit_error(data, 0, 2)
                else:
                    sid_by_accounts[data['accountId']] = sid
                    await sleep(0.2)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
            elif data['type'] == 'synchronize':
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                del sid_by_accounts[data['accountId']]
                await fake_server.respond(data)

        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 3})
            account2 = await api.metatrader_account_api.get_account('accountId2')
            connection2 = await account2.connect()
            await connection2.wait_synchronized({'timeoutInSeconds': 3})
            account3 = await api.metatrader_account_api.get_account('accountId3')
            connection3 = await account3.connect()
            await connection3.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] == sid_by_accounts['accountId2'] != sid_by_accounts['accountId3']
            await sleep(2)
            account4 = await api.metatrader_account_api.get_account('accountId4')
            connection4 = await account4.connect()
            await connection4.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] != sid_by_accounts['accountId4']
            await connection2.close()
            account5 = await api.metatrader_account_api.get_account('accountId5')
            connection5 = await account5.connect()
            await connection5.wait_synchronized({'timeoutInSeconds': 3})
            assert sid_by_accounts['accountId'] == sid_by_accounts['accountId5']

    @pytest.mark.asyncio
    async def test_resubscribe_on_disconnected_packet(self):
        """Should attempt to resubscribe on disconnected packet."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            assert connection.synchronized
            assert connection.terminal_state.connected
            assert connection.terminal_state.connected_to_broker
            fake_server.delete_status_task('accountId')
            await sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId', 'host': 'ps-mpa-0',
                                               'instanceIndex': 0})
            await sleep(0.2)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker
            await sleep(0.4)
            assert connection.synchronized
            assert connection.terminal_state.connected
            assert connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_multiple_streams(self):
        """Should handle multiple streams in one instance number."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            subscribe_called = False

            @sio.on('request')
            async def on_request(sid, data):
                if data['type'] == 'subscribe':
                    nonlocal subscribe_called
                    subscribe_called = True
                elif data['type'] == 'synchronize':
                    await fake_server.respond(data)
                    await fake_server.sync_account(data, 'ps-mpa-1')
                elif data['type'] == 'waitSynchronized':
                    await fake_server.respond(data)
                elif data['type'] == 'getAccountInformation':
                    await fake_server.respond_account_information(data)

            asyncio.create_task(fake_server.create_status_task('accountId', 'ps-mpa-1'))
            await fake_server.authenticate({'accountId': 'accountId'}, 'ps-mpa-1')
            await sleep(0.4)
            fake_server.delete_status_task('accountId')
            await sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId', 'host': 'ps-mpa-0',
                                               'instanceIndex': 0})
            await sleep(0.2)
            assert connection.synchronized
            assert connection.terminal_state.connected
            assert connection.terminal_state.connected_to_broker
            assert not subscribe_called
            await sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId', 'host': 'ps-mpa-1',
                                               'instanceIndex': 0})
            await sleep(0.1)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker

    @pytest.mark.asyncio
    async def test_multiple_streams_timeout(self):
        """Should not resubscribe if multiple streams and one timed out."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            subscribe_called = False

            @sio.on('request')
            async def on_request(sid, data):
                if data['type'] == 'subscribe':
                    nonlocal subscribe_called
                    subscribe_called = True
                elif data['type'] == 'synchronize':
                    await fake_server.respond(data)
                    await fake_server.sync_account(data, 'ps-mpa-1')
                elif data['type'] == 'waitSynchronized':
                    await fake_server.respond(data)
                elif data['type'] == 'getAccountInformation':
                    await fake_server.respond_account_information(data)

            status_task = asyncio.create_task(fake_server.create_status_task('accountId', 'ps-mpa-1'))
            await fake_server.authenticate({'accountId': 'accountId'}, 'ps-mpa-1')
            await sleep(0.1)
            fake_server.delete_status_task('accountId')
            await sleep(1.1)
            assert connection.synchronized
            assert connection.terminal_state.connected
            assert connection.terminal_state.connected_to_broker
            assert not subscribe_called
            status_task.cancel()
            await sleep(1.1)
            assert not connection.synchronized
            assert not connection.terminal_state.connected
            assert not connection.terminal_state.connected_to_broker
            assert subscribe_called

    @pytest.mark.asyncio
    async def test_not_resubscribe_after_close(self):
        """Should not synchronize if connection is closed."""
        synchronize_counter = 0

        @sio.on('request')
        async def on_request(sid, data):
            if data['type'] == 'subscribe':
                await sleep(0.2)
                await fake_server.respond(data)
                fake_server.status_tasks[data['accountId']] = \
                    asyncio.create_task(fake_server.create_status_task(data['accountId']))
                await fake_server.authenticate(data)
            elif data['type'] == 'synchronize':
                nonlocal synchronize_counter
                synchronize_counter += 1
                await fake_server.respond(data)
                await fake_server.sync_account(data)
            elif data['type'] == 'waitSynchronized':
                await fake_server.respond(data)
            elif data['type'] == 'getAccountInformation':
                await fake_server.respond_account_information(data)
            elif data['type'] == 'unsubscribe':
                await fake_server.respond(data)

        account = await api.metatrader_account_api.get_account('accountId')
        connection = await account.connect()
        await connection.wait_synchronized({'timeoutInSeconds': 3})
        assert synchronize_counter == 1
        account2 = await api.metatrader_account_api.get_account('accountId2')
        connection2 = await account2.connect()
        await sleep(0.1)
        await connection2.close()
        try:
            await connection2.wait_synchronized({'timeoutInSeconds': 3})
            raise Exception('TimeoutException expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
        assert synchronize_counter == 1

    @pytest.mark.asyncio
    async def test_not_resubscribe_after_close_after_disconnected(self):
        """Should not resubscribe after connection is closed."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 5)):
            subscribe_counter = 0

            @sio.on('request')
            async def on_request(sid, data):
                if data['type'] == 'subscribe':
                    nonlocal subscribe_counter
                    subscribe_counter += 1
                    await sleep(0.1)
                    await fake_server.respond(data)
                    fake_server.delete_status_task(data['accountId'])
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
                elif data['type'] == 'synchronize':
                    await fake_server.respond(data)
                    await fake_server.sync_account(data)
                elif data['type'] == 'waitSynchronized':
                    await fake_server.respond(data)
                elif data['type'] == 'getAccountInformation':
                    await fake_server.respond_account_information(data)
                elif data['type'] == 'unsubscribe':
                    fake_server.delete_status_task(data['accountId'])
                    await fake_server.respond(data)

            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker
            assert subscribe_counter == 1
            asyncio.create_task(sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId',
                                                             'host': 'ps-mpa-0', 'instanceIndex': 0}))
            await sleep(2)
            assert subscribe_counter > 1
            previous_subscribe_counter = subscribe_counter
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker
            asyncio.create_task(sio.emit('synchronization', {'type': 'disconnected', 'accountId': 'accountId',
                                                             'host': 'ps-mpa-0', 'instanceIndex': 0}))
            await sleep(0.1)
            await connection.close()
            await sleep(2)
            assert subscribe_counter == previous_subscribe_counter
            assert not connection.synchronized
            assert not connection.terminal_state.connected

    @pytest.mark.asyncio
    async def test_not_resubscribe_on_timeout_if_connection_closed(self):
        """Should not resubscribe on timeout if connection is closed."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            fake_server.status_tasks['accountId'].cancel()
            assert connection.synchronized
            await connection.close()
            await sleep(1.5)
            assert not connection.synchronized

    @pytest.mark.asyncio
    async def test_not_send_multiple_subscribes_on_status(self):
        """Should not send multiple subscribe requests if status arrives faster than subscribe."""
        with patch('lib.clients.metaApi.metaApiWebsocket_client.asyncio.sleep', new=lambda x: sleep(x / 50)):
            subscribe_calls = 0
            account = await api.metatrader_account_api.get_account('accountId')
            connection = await account.connect()
            await connection.wait_synchronized({'timeoutInSeconds': 10})
            fake_server.disable_sync()
            fake_server.status_tasks['accountId'].cancel()
            await sleep(2)
            assert not connection.synchronized and not connection.terminal_state.connected and \
                not connection.terminal_state.connected_to_broker

            @sio.on('request')
            async def on_request(sid, data):
                if data['type'] == 'subscribe':
                    nonlocal subscribe_calls
                    subscribe_calls += 1
                    await sleep(2.8)
                    await fake_server.respond(data)
                    fake_server.status_tasks[data['accountId']] = \
                        asyncio.create_task(fake_server.create_status_task(data['accountId']))
                    await fake_server.authenticate(data)
                elif data['type'] == 'synchronize':
                    await fake_server.respond(data)
                    await fake_server.sync_account(data)
                elif data['type'] == 'waitSynchronized':
                    await fake_server.respond(data)
                elif data['type'] == 'getAccountInformation':
                    await fake_server.respond_account_information(data)
                elif data['type'] == 'unsubscribe':
                    await fake_server.respond(data)

            fake_server.status_tasks['accountId'] = \
                asyncio.create_task(fake_server.create_status_task('accountId'))
            await sleep(4)
            assert connection.synchronized and connection.terminal_state.connected and \
                   connection.terminal_state.connected_to_broker
            assert subscribe_calls == 1



import os
import asyncio
from metaapi_cloud_sdk import MetaApi
from metaapi_cloud_sdk.clients.metaApi.synchronizationListener import SynchronizationListener
from metaapi_cloud_sdk.metaApi.models import MetatraderSymbolPrice, MetatraderCandle, MetatraderTick, MetatraderBook, \
    MarketDataSubscription, MarketDataUnsubscription
from typing import List

token = os.getenv('TOKEN') or '<put in your token here>'
account_id = os.getenv('ACCOUNT_ID') or '<put in your account id here>'
symbol = os.getenv('SYMBOL') or 'EURUSD'
domain = os.getenv('DOMAIN') or 'agiliumtrade.agiliumtrade.ai'


class QuoteListener(SynchronizationListener):
    async def on_symbol_price_updated(self, instance_index: int, price: MetatraderSymbolPrice):
        if price['symbol'] == symbol:
            print(symbol + ' price updated', price)

    async def on_candles_updated(self, instance_index: int, candles: List[MetatraderCandle], equity: float = None,
                                 margin: float = None, free_margin: float = None, margin_level: float = None,
                                 account_currency_exchange_rate: float = None):
        for candle in candles:
            if candle['symbol'] == symbol:
                print(symbol + ' candle updated', candle)

    async def on_ticks_updated(self, instance_index: int, ticks: List[MetatraderTick], equity: float = None,
                               margin: float = None, free_margin: float = None, margin_level: float = None,
                               account_currency_exchange_rate: float = None):
        for tick in ticks:
            if tick['symbol'] == symbol:
                print(symbol + ' tick updated', tick)

    async def on_books_updated(self, instance_index: int, books: List[MetatraderBook], equity: float = None,
                               margin: float = None, free_margin: float = None, margin_level: float = None,
                               account_currency_exchange_rate: float = None):
        for book in books:
            if book['symbol'] == symbol:
                print(symbol + ' order book updated', book)

    async def on_subscription_downgraded(self, instance_index: int, symbol: str,
                                         updates: List[MarketDataSubscription] or None = None,
                                         unsubscriptions: List[MarketDataUnsubscription] or None = None):
        print('Market data subscriptions for ' + symbol + ' were downgraded by the server due to rate limits')


async def stream_quotes():
    api = MetaApi(token, {'domain': domain})
    try:
        account = await api.metatrader_account_api.get_account(account_id)

        #  wait until account is deployed and connected to broker
        print('Deploying account')
        if account.state != 'DEPLOYED':
            await account.deploy()
        else:
            print('Account already deployed')
        print('Waiting for API server to connect to broker (may take couple of minutes)')
        if account.connection_status != 'CONNECTED':
            await account.wait_connected()

        # connect to MetaApi API
        connection = await account.connect()

        quote_listener = QuoteListener()
        connection.add_synchronization_listener(quote_listener)

        # wait until terminal state synchronized to the local state
        print('Waiting for SDK to synchronize to terminal state (may take some time depending on your history size), the price streaming will start once synchronization finishes')
        await connection.wait_synchronized()

        # Add symbol to MarketWatch if not yet added and subscribe to market data
        # Please note that currently only G1 and MT4 G2 instances support extended subscription management
        # Other instances will only stream quotes in response
        # Market depth streaming is available in MT5 only
        # Ticks streaming is not available for MT4 G1
        await connection.subscribe_to_market_data(symbol, [
            {'type': 'quotes', 'intervalInMilliseconds': 5000},
            {'type': 'candles', 'timeframe': '1m', 'intervalInMilliseconds': 10000},
            {'type': 'ticks'},
            {'type': 'marketDepth', 'intervalInMilliseconds': 5000}
        ])

        print(f'[{datetime.now().isoformat()}] Synchronized successfully, streaming ' + symbol + ' market data now...')

        while True:
            await asyncio.sleep(1)

    except Exception as err:
        print(api.format_error(err))

asyncio.run(stream_quotes())



from ..metaApi_client import MetaApiClient
from typing_extensions import TypedDict
from typing import List, Union, Optional, Dict
from httpx import Response
from enum import Enum


class Extension(TypedDict):
    """Extension model."""
    id: str
    """Extension id."""
    configuration: Dict
    """Extension configuration."""


class State(Enum):
    """Account state."""
    CREATED = 'CREATED'
    DEPLOYING = 'DEPLOYING'
    DEPLOYED = 'DEPLOYED'
    DEPLOY_FAILED = 'DEPLOY_FAILED'
    UNDEPLOYING = 'UNDEPLOYING'
    UNDEPLOYED = 'UNDEPLOYED'
    UNDEPLOY_FAILED = 'UNDEPLOY_FAILED'
    DELETING = 'DELETING'
    DELETE_FAILED = 'DELETE_FAILED'
    REDEPLOY_FAILED = 'REDEPLOY_FAILED'


class ConnectionStatus(Enum):
    """Account connection status."""
    CONNECTED = 'CONNECTED'
    DISCONNECTED = 'DISCONNECTED'
    DISCONNECTED_FROM_BROKER = 'DISCONNECTED_FROM_BROKER'


class AccountsFilter(TypedDict):

    offset: Optional[int]
    """Search offset (defaults to 0) (must be greater or equal to 0)."""
    limit: Optional[int]
    """Search limit (defaults to 1000) (must be greater or equal to 1 and less or equal to 1000)."""
    version: Optional[Union[List[int], int]]
    """MT version (allowed values are 4 and 5)"""
    type: Optional[Union[List[str], str]]
    """Account type. Allowed values are 'cloud' and 'self-hosted'"""
    state: Optional[Union[List[State], State]]
    """Account state."""
    connectionStatus: Optional[Union[List[ConnectionStatus], ConnectionStatus]]
    """Connection status."""
    query: Optional[str]
    """Searches over _id, name, server and login to match query."""
    provisioningProfileId: Optional[str]
    """Provisioning profile id."""


class MetatraderAccountIdDto(TypedDict):
    """MetaTrader account id model"""

    id: str
    """MetaTrader account unique identifier"""


class MetatraderAccountDto(TypedDict):
    """MetaTrader account model"""

    _id: str
    """Account unique identifier."""
    name: str
    """MetaTrader account human-readable name in the MetaApi app."""
    type: str
    """Account type, can be cloud, cloud-g1, cloud-g2 or self-hosted. Cloud and cloud-g2 are aliases."""
    login: str
    """MetaTrader account number."""
    server: str
    """MetaTrader server which hosts the account."""
    version: int
    """MT version (allowed values are 4 and 5)."""
    provisioningProfileId: str
    """Id of the account's provisioning profile."""
    application: str
    """Application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade"""
    magic: int
    """MetaTrader magic to place trades using."""
    state: str
    """Account deployment state. One of CREATED, DEPLOYING, DEPLOYED, UNDEPLOYING, UNDEPLOYED, DELETING"""
    connectionStatus: str
    """Terminal & broker connection status, one of CONNECTED, DISCONNECTED, DISCONNECTED_FROM_BROKER"""
    accessToken: str
    """Authorization token to be used for accessing single account data. Intended to be used in browser API."""
    manualTrades: bool
    """Flag indicating if trades should be placed as manual trades. Default is false."""
    quoteStreamingIntervalInSeconds: float
    """Quote streaming interval in seconds. Set to 0 in order to receive quotes on each tick. Default value is
    2.5 seconds. Intervals less than 2.5 seconds are supported only for G2."""
    tags: Optional[List[str]]
    """MetaTrader account tags."""
    extensions: List[Extension]
    """API extensions."""
    metadata: Dict
    """Extra information which can be stored together with your account."""
    reliability: str
    """Used to increase the reliability of the account. Allowed values are regular and high. Default is regular."""
    baseCurrency: str
    """3-character ISO currency code of the account base currency. Default value is USD.
    The setting is to be used for copy trading accounts which use national currencies only, such as some Brazilian
    brokers. You should not alter this setting unless you understand what you are doing."""
    copyFactoryRoles: List[str]
    """Account roles for CopyFactory2 application. Allowed values are `PROVIDER` and `SUBSCRIBER`."""
    resourceSlots: int
    """Number of resource slots to allocate to account. Allocating extra resource slots results in better account
    performance under load which is useful for some applications. E.g. if you have many accounts copying the same
    strategy via CopyFactory API, then you can increase resourceSlots to get a lower trade copying latency. Please
    note that allocating extra resource slots is a paid option. Default is 1."""


class NewMetatraderAccountDto(TypedDict):
    """New MetaTrader account model"""

    name: str
    """MetaTrader account human-readable name in the MetaApi app."""
    type: str
    """Account type, can be cloud, cloud-g1, cloud-g2 or self-hosted. cloud-g2 and cloud are aliases. When you
    create MT5 cloud account the type is automatically converted to cloud-g1 because MT5 G2 support is still
    experimental. You can still create MT5 G2 account by setting type to cloud-g2."""
    login: str
    """MetaTrader account number."""
    password: str
    """MetaTrader account password. The password can be either investor password for read-only access or master
    password to enable trading features. Required for cloud account."""
    server: str
    """MetaTrader server which hosts the account."""
    provisioningProfileId: str
    """Id of the account's provisioning profile."""
    application: str
    """Application name to connect the account to. Currently allowed values are MetaApi and AgiliumTrade."""
    magic: int
    """MetaTrader magic to place trades using."""
    manualTrades: bool
    """Flag indicating if trades should be placed as manual trades. Default is false."""
    quoteStreamingIntervalInSeconds: float
    """Quote streaming interval in seconds. Set to 0 in order to receive quotes on each tick. Default value is
    2.5 seconds. Intervals less than 2.5 seconds are supported only for G2."""
    tags: Optional[List[str]]
    """MetaTrader account tags."""
    extensions: List[Extension]
    """API extensions."""
    metadata: Dict
    """Extra information which can be stored together with your account."""
    reliability: str
    """Used to increase the reliability of the account. Allowed values are regular and high. Default is regular."""
    baseCurrency: str
    """3-character ISO currency code of the account base currency. Default value is USD.
    The setting is to be used for copy trading accounts which use national currencies only, such as some Brazilian
    brokers. You should not alter this setting unless you understand what you are doing."""
    copyFactoryRoles: List[str]
    """Account roles for CopyFactory2 application. Allowed values are `PROVIDER` and `SUBSCRIBER`."""
    resourceSlots: int
    """Number of resource slots to allocate to account. Allocating extra resource slots results in better account
    performance under load which is useful for some applications. E.g. if you have many accounts copying the same
    strategy via CopyFactory API, then you can increase resourceSlots to get a lower trade copying latency. Please
    note that allocating extra resource slots is a paid option. Default is 1."""


class MetatraderAccountUpdateDto(TypedDict):
    """Updated MetaTrader account data"""

    name: str
    """MetaTrader account human-readable name in the MetaApi app."""
    password: str
    """MetaTrader account password. The password can be either investor password for read-only
    access or master password to enable trading features. Required for cloud account"""
    server: str
    """MetaTrader server which hosts the account"""
    manualTrades: bool
    """Flag indicating if trades should be placed as manual trades. Default is false."""
    quoteStreamingIntervalInSeconds: float
    """Quote streaming interval in seconds. Set to 0 in order to receive quotes on each tick. Default value is
    2.5 seconds. Intervals less than 2.5 seconds are supported only for G2."""
    tags: Optional[List[str]]
    """MetaTrader account tags."""
    extensions: List[Extension]
    """API extensions."""
    metadata: Dict
    """Extra information which can be stored together with your account."""
    copyFactoryRoles: List[str]
    """Account roles for CopyFactory2 application. Allowed values are `PROVIDER` and `SUBSCRIBER`."""
    resourceSlots: int
    """Number of resource slots to allocate to account. Allocating extra resource slots results in better account
    performance under load which is useful for some applications. E.g. if you have many accounts copying the same
    strategy via CopyFactory API, then you can increase resourceSlots to get a lower trade copying latency. Please
    note that allocating extra resource slots is a paid option. Default is 1."""


class MetatraderAccountClient(MetaApiClient):
    """metaapi.cloud MetaTrader account API client (see https://metaapi.cloud/docs/provisioning/)

    Attributes:
        _httpClient: HTTP client
        _host: domain to connect to
        _token: authorization token
    """

    async def get_accounts(self, accounts_filter: AccountsFilter = None) -> Response:
        """Retrieves MetaTrader accounts owned by user
        (see https://metaapi.cloud/docs/provisioning/api/account/readAccounts/)

        Args:
            accounts_filter: Optional filter.

        Returns:
            A coroutine resolving with List[MetatraderAccountDto] - MetaTrader accounts found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_accounts')
        opts = {
            'url': f'{self._host}/users/current/accounts',
            'method': 'GET',
            'params': accounts_filter or {},
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_account(self, id: str) -> Response:
        """Retrieves a MetaTrader account by id (see https://metaapi.cloud/docs/provisioning/api/account/readAccount/).
        Throws an error if account is not found.

        Args:
            id: MetaTrader account id.

        Returns:
            A coroutine resolving with MetatraderAccountDto - MetaTrader account found.
        """
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}',
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_account_by_token(self) -> 'Response[MetatraderAccountDto]':
        """Retrieves a MetaTrader account by token
        (see https://metaapi.cloud/docs/provisioning/api/account/readAccount/). Throws an error if account is
        not found. Method is accessible only with account access token.

        Returns:
            A coroutine resolving with MetaTrader account found.
        """
        if self._is_not_account_token():
            return self._handle_no_access_exception('get_account_by_token')
        opts = {
            'url': f'{self._host}/users/current/accounts/accessToken/{self._token}',
            'method': 'GET'
        }
        return await self._httpClient.request(opts)

    async def create_account(self, account: NewMetatraderAccountDto) -> Response:
        """Starts cloud API server for a MetaTrader account using specified provisioning profile
        (see https://metaapi.cloud/docs/provisioning/api/account/createAccount/).
        It takes some time to launch the terminal and connect the terminal to the broker, you can use the
        connectionStatus field to monitor the current status of the terminal.

        Args:
            account: MetaTrader account to create.

        Returns:
            A coroutine resolving with MetatraderAccountIdDto - an id of the MetaTrader account created.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('create_account')
        opts = {
            'url': f'{self._host}/users/current/accounts',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)

    async def deploy_account(self, id: str) -> Response:
        """Starts API server for MetaTrader account. This request will be ignored if the account has already
        been deployed. (see https://metaapi.cloud/docs/provisioning/api/account/deployAccount/)

        Args:
            id: MetaTrader account id to deploy.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for deployment
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('deploy_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/deploy',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def undeploy_account(self, id: str) -> Response:
        """Stops API server for a MetaTrader account. Terminal data such as downloaded market history data will
        be preserved. (see https://metaapi.cloud/docs/provisioning/api/account/undeployAccount/)

        Args:
            id: MetaTrader account id to undeploy.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for undeployment.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('undeploy_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/undeploy',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def redeploy_account(self, id: str) -> Response:
        """Redeploys MetaTrader account. This is equivalent to undeploy immediately followed by deploy.
        (see https://metaapi.cloud/docs/provisioning/api/account/redeployAccount/)

        Args:
            id: MetaTrader account id to redeploy.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for redeployment.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('redeploy_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/redeploy',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def delete_account(self, id: str) -> Response:
        """Stops and deletes an API server for a specified MetaTrader account. The terminal state such as downloaded
        market data history will be deleted as well when you delete the account.
        (see https://metaapi.cloud/docs/provisioning/api/account/deleteAccount/)

        Args:
            id: MetaTrader account id to delete.

        Returns:
            A coroutine resolving when MetaTrader account is scheduled for deletion.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('delete_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}',
            'method': 'DELETE',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def update_account(self, id: str, account: MetatraderAccountUpdateDto) -> Response:
        """Updates existing metatrader account data (see
        https://metaapi.cloud/docs/provisioning/api/account/updateAccount/)

        Args:
            id: MetaTrader account id.
            account: Updated MetaTrader account.

        Returns:
            A coroutine resolving when MetaTrader account is updated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('update_account')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}',
            'method': 'PUT',
            'headers': {
                'auth-token': self._token
            },
            'body': account
        }
        return await self._httpClient.request(opts)

    async def increase_reliability(self, id: str):
        """Increases MetaTrader account reliability. The account will be temporary stopped to perform this action. (see
        https://metaapi.cloud/docs/provisioning/api/account/increaseReliability/).
        Method is accessible only with API access token.

        Args:
            id: MetaTrader account id.

        Returns:
            A coroutine resolving when MetaTrader account reliability is increased.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('increase_reliability')
        opts = {
            'url': f'{self._host}/users/current/accounts/{id}/increase-reliability',
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)
